/*
 * O2xIV
 * Copyright (C) 2008 Carl Reinke
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
#include "files.h"
#include "font.h"
#include "input.h"
#include "main.h"
#include "menu.h"
#include "prefs.h"
#include "scale.h"
#include "thread.h"

#include "SDL.h"

#include <assert.h>
#include <libgen.h>
#include <stdbool.h>
#include <string.h>
#include <sys/stat.h>

#ifndef strdup
#define strdup(src) (strcpy(malloc(strlen(src) + 1), src))
#endif

int main( int argc, char *argv[] );

void set_rotation( int new_rotation );
void calculate_auto_fit( void );

void redraw( void );
void rotate( void );

bool quit = false;

bool need_redraw = true;
int delay;

#define NOTE_SHOW  0x01
#define NOTE_DELAY 0x02
#define NOTE_MENU  0x04
int note_flags[3] = { NOTE_MENU, 0, NOTE_MENU };
Uint32 note_delay[3];
char note[3][50];

SDL_Surface *surface, *image, *scaled, *scaled_rot[4];

fixed fit_outside, fit_inside;

bool reset_top_once = false;

fixed scale = int_to_fixed(1),
      scale_inv = int_to_fixed(1); // multiplicative inverse
int rotation = 0; // 0 is 0°, 1 is 90°, 2 is 180°, 3 is 270°

SDL_Rect pan_rect, drawn_rect, scaled_rect, rotated_rect, surface_rect;

SDL_VideoInfo *video_info = NULL;;

int 	SCREEN_WIDTH	= 320;
int 	SCREEN_HEIGHT	= 240;
int 	SCREEN_DEPTH	= 16;
int		SCREEN_BPP		= 2;
Uint32	SCREEN_MODE		= SDL_HWSURFACE | SDL_DOUBLEBUF;

int main( int argc, char *argv[] ) {
	if (SDL_Init(SDL_INIT_TIMER | SDL_INIT_VIDEO) == -1) {
		printf("%s:%d\n", __FILE__, __LINE__);
		return -1;
	}

	video_info = SDL_GetVideoInfo();

#ifdef TARGET_GP2X
	SCREEN_WIDTH	= 320;
	SCREEN_HEIGHT	= 240;
	SCREEN_DEPTH	= 16;
	SCREEN_MODE		= SDL_HWSURFACE | SDL_DOUBLEBUF;
#else
	SCREEN_WIDTH	= video_info->current_w;
	SCREEN_HEIGHT	= video_info->current_h;
	//SCREEN_DEPTH	= video_info->vfmt->BitsPerPixel;
	//SCREEN_BPP		= video_info->vfmt->BytesPerPixel;
	SCREEN_MODE		|= SDL_FULLSCREEN;
//	SCREEN_WIDTH	= 1920;
//	SCREEN_HEIGHT	= 1200;
//	SCREEN_DEPTH	= 16;
#endif



	surface = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_DEPTH, SCREEN_MODE);
	if (!surface) {
		printf("%s:%d\n", __FILE__, __LINE__);
		return -1;
	}



#ifdef TARGET_GP2X
	SDL_ShowCursor(0);
#endif /* TARGET_GP2X */
	SDL_ShowCursor(0);

	// must be 16 bits per pixel or the scaler will trhow an assertion.
	scaled_rot[0] = SDL_CreateRGBSurfaceFrom(malloc(SCREEN_WIDTH * SCREEN_HEIGHT * SCREEN_BPP), SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_DEPTH, SCREEN_WIDTH * 2, 0, 0, 0, 0);
	scaled_rot[1] = SDL_CreateRGBSurfaceFrom(malloc(SCREEN_HEIGHT * SCREEN_WIDTH * SCREEN_BPP), SCREEN_HEIGHT, SCREEN_WIDTH, SCREEN_DEPTH, SCREEN_HEIGHT * 2, 0, 0, 0, 0);
	scaled_rot[2] = SDL_CreateRGBSurfaceFrom(malloc(SCREEN_WIDTH * SCREEN_HEIGHT * SCREEN_BPP), SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_DEPTH, SCREEN_WIDTH * 2, 0, 0, 0, 0);
	scaled_rot[3] = scaled_rot[1];

	init_input();

	// parse parameter (file or directory)
	if (argc > 1) {
		struct stat temp;
		stat(argv[1], &temp);

		if (S_ISREG(temp.st_mode) && maybe_image(argv[1])) {
			char *path_temp = strdup(argv[1]);
			path = strdup(dirname(path_temp));
			if (strcmp(path, "/") != 0)
				strcat(path, "/");
			free(path_temp);

			populate_files(path, true);

			path_temp = strdup(argv[1]);
			char *filename = basename(path_temp);

			for (file_index = file_min; file_index < file_max; file_index++)
				if (strcmp(filename, files[file_index]) == 0)
					break;
			assert(file_index < file_max);

			free(path_temp);
		} else if (S_ISDIR(temp.st_mode)) {
			bool slash = argv[1][strlen(argv[1]) - 1] != '/';

			path = malloc(strlen(argv[1]) + 3 + 1);
			sprintf(path, "%s%s", argv[1], slash ? "/" : "");

			populate_files(path, true);

			toggle_file_selector();
		}
	}

	if (path == NULL) {
		path = malloc(100); // arbitrary initial value, just don't overflow it yet
#ifdef TARGET_GP2X
		strcpy(path, "/mnt/sd/");
#else
		strcpy(path, "/");
#endif

		populate_files(path, true);

		toggle_file_selector();
	}

	image_modify_mutex = SDL_CreateMutex();
	files_modify_mutex = SDL_CreateMutex();
	SDL_Thread *load_images_thread = SDL_CreateThread(load_images, NULL);

	SDL_LockMutex(image_modify_mutex);

	load_prefs();

	while (!quit) {	/* Main loop */
		delay = SDL_GetTicks() + 20;

		poll_input();

		// handle input
		if (file_selector) {
			file_selector_input();
		} else if (menu) {
			menu_input();
		} else { // viewing input
			switch (last_key) {
				case SDLK_ESCAPE:
					quit = true;
					break;

				case GP2X_KB_SELECT:
					toggle_file_selector();
					need_redraw = true;
					break;

				case GP2X_KB_START:
					toggle_menu();
					need_redraw = true;
					break;

				case GP2X_KB_Y:
				case GP2X_KB_X:
				{
					fixed ratio = last_key == GP2X_KB_Y ? int_div_int_to_fixed(5, 4) : int_div_int_to_fixed(4, 5);
					set_scale(fixed_mul(scale, ratio));
					need_redraw = true;
					break;
				}

				case GP2X_KB_A:
				case GP2X_KB_B:
					set_rotation(rotation + (last_key == GP2X_KB_A ? 3 : 1));
					set_scale(auto_fit ? (fit_pref == 1 ? fit_outside : fit_inside) : scale);
					need_redraw = true;
					break;

				case GP2X_KB_L:
				case GP2X_KB_R:
					file_index = wrap_file_index(file_index + (last_key == GP2X_KB_L ? -1 : 1), true);
					input[0] = input[1] = input[2] = input[3] = 0; // Fixs a bug that happens when scrolling and changing the image at the same time
					break;

				default:
					break;
			}

			const int scroll = 20;

			if (input[0] || input[1] || input[2] || input[3]) {
				pan_rect.x += scroll * (input[(1 + rotation) % 4] - input[(3 + rotation) % 4]);
				pan_rect.y += scroll * (input[(2 + rotation) % 4] - input[(0 + rotation) % 4]);

				need_redraw = true;
			}
			if (input[4])
				need_redraw = true;
		}
		last_key = 0;

		// hide notes on timeout
		for (unsigned int i = 0; i < countof(note); i++)
			if (note_flags[i] & NOTE_DELAY && note_delay[i] < SDL_GetTicks() && note_delay[i]) {
				note_flags[i] &= ~NOTE_DELAY;
				note_delay[i] = 0;

				need_redraw = true;
			}

		if (need_redraw) {
			need_redraw = false;

			if (image)
				redraw();
			else
				SDL_FillRect(surface, NULL, 0);

			if SDL_MUSTLOCK(surface) SDL_LockSurface(surface);

			// display notes
			for (unsigned int i = 0; i < countof(note_delay); i++) {
				if (note_flags[i] & (NOTE_SHOW | NOTE_DELAY | (menu ? NOTE_MENU : 0)) || note_delay[i]) {
					int x = i & 1 ? 2 : SCREEN_WIDTH - strlen(shorten(note[i], 39)) * 8 - 2,
						y = i > 0 ? SCREEN_HEIGHT - 8 - 2 : 2;
					outline8x8(surface, x, y, shorten(note[i], 39), 0xffff, 0x0000);

					if (note_flags[i] & NOTE_DELAY && note_delay[i] == 0)
						note_delay[i] = SDL_GetTicks() + 1000;
				}
			}

			if SDL_MUSTLOCK(surface) SDL_UnlockSurface(surface);

			if (file_selector) {
				draw_file_selector();
			} else if (menu) {
				draw_menu();
			}

			if SDL_MUSTLOCK(surface) SDL_LockSurface(surface);

			if (input[4]) {
				outline8x8(surface, 4, SCREEN_HEIGHT - 10, "Open2xIV v0.10", 0xffff, 0x0000);

				const char buffer[] = "!sseldniM yb dedoc"; // sneaky... not really
				static int j = 0; j++;
				for (unsigned int i = 0; i < strlen(buffer); i++) {
					char8x8(surface, SCREEN_WIDTH - i * 8 - 12 - 1, SCREEN_HEIGHT - 10,     buffer[i], 0xffff);
					char8x8(surface, SCREEN_WIDTH - i * 8 - 12,     SCREEN_HEIGHT - 10 - 1, buffer[i], 0xffff);
					char8x8(surface, SCREEN_WIDTH - i * 8 - 12 + 1, SCREEN_HEIGHT - 10,     buffer[i], 0xffff);
					char8x8(surface, SCREEN_WIDTH - i * 8 - 12,     SCREEN_HEIGHT - 10 + 1, buffer[i], 0xffff);
					char8x8(surface, SCREEN_WIDTH - i * 8 - 12,     SCREEN_HEIGHT - 10,     buffer[i], (i + j) % 3 == 2 ? 0xd000 : ((i + j) % 3 == 1 ? 0x0340 : 0x001a));
				}

				delay += 100;
				need_redraw = true;
			}

			if SDL_MUSTLOCK(surface) SDL_UnlockSurface(surface);

			//static int j = 0; char buffer[8]; sprintf(buffer, "%d", j++); text8x8(surface, 0, 0, buffer, 0xffff); // frame counter

			SDL_Flip(surface);
		}

		SDL_UnlockMutex(image_modify_mutex);

		delay -= SDL_GetTicks();
		if (delay > 0)
			SDL_Delay(delay);

		SDL_LockMutex(image_modify_mutex);

		if (image_update) {
			if (image_loaded) {
				sprintf(note[0], "%s %d/%d", files[file_index], file_index - file_min + 1, file_max - file_min);
				note_flags[0] |= NOTE_DELAY;
				note_delay[0] = 0;

				note_flags[1] &= ~NOTE_SHOW;
			}

			if (image_loading) {
				strcpy(note[1], "Loading...");
				note_flags[1] |= NOTE_SHOW;
			}

			if (image_loaded && image == NULL) {
				note_flags[0] |= NOTE_SHOW;

				strcpy(note[1], "Nothing to display!");
				note_flags[1] |= NOTE_SHOW;

				strcpy(note[2], "");
			} else {
				note_flags[0] &= ~NOTE_SHOW;
			}

			if (image_loaded && image != NULL) {
				set_rotation(auto_rotate && image->h > image->w ? 3 : 0);
				set_scale(auto_fit ? (fit_pref == 1 ? fit_outside : fit_inside) : scale);
			}

			image_update = false;
			need_redraw = true;
		}

	}

	save_prefs();

	if (file_selector)
		SDL_UnlockMutex(files_modify_mutex);

	SDL_UnlockMutex(image_modify_mutex);

	thread_die = true;
	SDL_WaitThread(load_images_thread, NULL);

	SDL_DestroyMutex(files_modify_mutex);
	SDL_DestroyMutex(image_modify_mutex);

	free(scaled_rot[0]->pixels); // yeah, this is wrong, but I'm lazy
	SDL_FreeSurface(scaled_rot[0]);
	free(scaled_rot[1]->pixels);
	SDL_FreeSurface(scaled_rot[1]);
	free(scaled_rot[2]->pixels);
	SDL_FreeSurface(scaled_rot[2]);
	SDL_FreeSurface(surface);

	SDL_Quit();

	return 0;
}

void set_rotation( int new_rotation ) {
	if (!image_loaded)
		return;

	rotation = new_rotation % 4;
	scaled = scaled_rot[rotation];

	calculate_auto_fit();
}

void rotate( void ) {
	const int Bpp = 2;

	Uint8 *src = scaled->pixels,
	      *dst = scaled_rot[0]->pixels;

	if (rotation == 0) {
		return;
	} else if (rotation == 2) {
		const int src_w = SCREEN_WIDTH, src_h = SCREEN_HEIGHT;

		const int dst_pitch = SCREEN_WIDTH * Bpp;

		dst += src_h /*dst_h*/ * dst_pitch;
		for (int y = src_h; y; y--) {
			// dst -= dst_extra;
			for (int x = src_w; x; x--) {
				dst -= Bpp;
				*(Uint16 *)dst = *(Uint16 *)src;
				src += Bpp;
			}
			// src += src_extra;
		}
		return;
	} else {
		const int src_w = SCREEN_HEIGHT, src_h = SCREEN_WIDTH,
		          dst_w = SCREEN_WIDTH, dst_h = SCREEN_HEIGHT;

		const int dst_pitch = SCREEN_WIDTH * Bpp;

		switch (rotation)
		{
			case 1:
				dst += dst_w * Bpp;
				for (int y = src_h; y; y--) {
					dst -= Bpp;
					for (int x = src_w; x; x--) {
						*(Uint16 *)dst = *(Uint16 *)src;
						src += Bpp;
						dst += dst_pitch;
					}
					// src += src_extra;
					dst -= dst_h * dst_pitch;
				}
				return;
			case 3:
				for (int y = src_h; y; y--) {
					dst += dst_h * dst_pitch;
					for (int x = src_w; x; x--) {
						dst -= dst_pitch;
						*(Uint16 *)dst = *(Uint16 *)src;
						src += Bpp;
					}
					// src += src_extra;
					dst += Bpp;
				}
				return;
			default:
				assert(false);
				return;
		}
	}
}

void calculate_auto_fit( void ) {
	fit_outside = int_to_fixed(scaled->w) / image->w;
	fit_inside  = int_to_fixed(scaled->h) / image->h;
	if (fit_outside < fit_inside) {
		fixed temp = fit_outside;
		fit_outside = fit_inside;
		fit_inside = temp;
	}
}

void set_scale( fixed new_scale ) {
	if (!image_loaded)
		return;

	new_scale = min(max(new_scale, int_div_int_to_fixed(1, 100)), int_to_fixed(4));

	// prepare for recentering
	int center_x = 0, center_y = 0;
	if (scaled_rect.w != 0) {
		center_x = fixed_mul((pan_rect.x + scaled_rect.w / 2), fixed_div(new_scale, scale));
		center_y = fixed_mul((pan_rect.y + scaled_rect.h / 2), fixed_div(new_scale, scale));
	}

	if (scale != new_scale) {
		if (scale < (int_div_int_to_fixed(1, 10)))
			sprintf(note[2], "%ix%i %.1f%%", image->w, image->h, fixed_to_float(new_scale * 100));
		else
			sprintf(note[2], "%ix%i %.0f%%", image->w, image->h, fixed_to_float(new_scale * 100));
		note_flags[2] |= NOTE_DELAY;
		note_delay[2] = 0;
	}

	scale = new_scale;
	scale_inv = int_div_fixed_to_fixed(1, scale);

	// panning maximums (dimensions of scaled image)
	pan_rect.w = fixed_mul(image->w, scale);
	pan_rect.h = fixed_mul(image->h, scale);

	// for blitting from scaled surface, which is max of SCREEN_WIDTHxSCREEN_HEIGHT
	scaled_rect.w = min(pan_rect.w, scaled->w);
	scaled_rect.h = min(pan_rect.h, scaled->h);

	// recenter the panning when image larger than SCREEN_WIDTHxSCREEN_HEIGHT
	pan_rect.x = max(center_x - scaled_rect.w / 2, 0);
	pan_rect.y = max(center_y - scaled_rect.h / 2, 0);

	drawn_rect.w = drawn_rect.h = 0;

	// printf("%dx%d\t%3.1f%%\t%d°\t%dx%d\n", image->w, image->h, fixed_to_float(scale * 100), rotation * 90, pan_rect.w, pan_rect.h);
}

void redraw( void ) {
	void (*fixed_scale)( SDL_Surface *src, SDL_Surface *dst, const fixed src_x, fixed src_y, const SDL_Rect *dst_area, const fixed scale, const fixed scale_inv ) = &fixed_scale_16;

	switch (image->format->BytesPerPixel) {
		case 4:
			fixed_scale = &fixed_scale_32;
			break;
		case 3:
			fixed_scale = &fixed_scale_24;
			break;
		case 2:
			fixed_scale = &fixed_scale_16;
			break;
		case 1:
			fixed_scale = &fixed_scale_8;
			break;
		default:
			assert(0);
			break;
	}

	// keep the panning within the image
	pan_rect.x = max(min(pan_rect.x, pan_rect.w - scaled_rect.w), 0);
	pan_rect.y = max(min(pan_rect.y, pan_rect.h - scaled_rect.h), 0);

	if (drawn_rect.w == 0) {
		//  printf("pan (scale)\tx %d\ty %d\tx %d\ty %d\n", pan_rect.x, pan_rect.y, pan_rect.x + min(pan_rect.w, scaled->w), pan_rect.y + min(pan_rect.h, scaled->h)); fflush(stdout);

		if SDL_MUSTLOCK(image) SDL_LockSurface(image);
		if SDL_MUSTLOCK(scaled) SDL_LockSurface(scaled);

		fixed_scale(image, scaled, pan_rect.x * scale_inv, pan_rect.y * scale_inv, &scaled_rect, scale, scale_inv);

		if SDL_MUSTLOCK(image) SDL_UnlockSurface(image);
		if SDL_MUSTLOCK(scaled) SDL_UnlockSurface(scaled);

	// scale partial screen
	} else if (pan_rect.x != drawn_rect.x || pan_rect.y != drawn_rect.y) {
		//  printf("pan (partial)\tx %d\ty %d\tx %d\ty %d\n", pan_rect.x, pan_rect.y, pan_rect.x + min(pan_rect.w, scaled->w), pan_rect.y + min(pan_rect.h, scaled->h)); fflush(stdout);

		// shift already drawn image
		SDL_Rect move_rect = { drawn_rect.x - pan_rect.x, drawn_rect.y - pan_rect.y, 0, 0 };
		// printf("move\tx %d\ty %d\tx %d\ty %d\n", move_rect.x, move_rect.y, move_rect.x + move_rect.w, move_rect.y + move_rect.h); fflush(stdout);
		SDL_BlitSurface(scaled, NULL, scaled, &move_rect);

		if SDL_MUSTLOCK(image) SDL_LockSurface(image);
		if SDL_MUSTLOCK(scaled) SDL_LockSurface(scaled);

		// scale the rest
		if (pan_rect.x < drawn_rect.x) { // left
			SDL_Rect draw_rect = { 0, 0, drawn_rect.x - pan_rect.x, scaled_rect.h };
#ifdef TARGET_GP2X
			draw_rect.w++; // cover over ugly SDL bug
#endif
			// printf("draw\tx %d\ty %d\tx %d\ty %d\n", draw_rect.x, draw_rect.y, draw_rect.x + draw_rect.w, draw_rect.y + draw_rect.h); fflush(stdout);
			fixed_scale(image, scaled, pan_rect.x * scale_inv, pan_rect.y * scale_inv, &draw_rect, scale, scale_inv);
		}
		if (pan_rect.x > drawn_rect.x) { // right
			SDL_Rect draw_rect = { scaled_rect.w - (pan_rect.x - drawn_rect.x), 0, pan_rect.x - drawn_rect.x, scaled_rect.h };
			// printf("draw\tx %d\ty %d\tx %d\ty %d\n", draw_rect.x, draw_rect.y, draw_rect.x + draw_rect.w, draw_rect.y + draw_rect.h); fflush(stdout);
			fixed_scale(image, scaled, (pan_rect.x + draw_rect.x) * scale_inv, pan_rect.y * scale_inv, &draw_rect, scale, scale_inv);
		}
		if (pan_rect.y < drawn_rect.y) { // top
			SDL_Rect draw_rect = { 0, 0, scaled_rect.w, drawn_rect.y - pan_rect.y };
			// printf("draw\tx %d\ty %d\tx %d\ty %d\n", draw_rect.x, draw_rect.y, draw_rect.x + draw_rect.w, draw_rect.y + draw_rect.h); fflush(stdout);
			fixed_scale(image, scaled, pan_rect.x * scale_inv, pan_rect.y * scale_inv, &draw_rect, scale, scale_inv);
#ifdef TARGET_GP2X
			draw_rect.h++; // cover over ugly SDL bug
#endif
		}
		if (pan_rect.y > drawn_rect.y) { // bottom
			SDL_Rect draw_rect = { 0, scaled_rect.h - (pan_rect.y - drawn_rect.y), scaled_rect.w, pan_rect.y - drawn_rect.y };
			// printf("draw\tx %d\ty %d\tx %d\ty %d\n", draw_rect.x, draw_rect.y, draw_rect.x + draw_rect.w, draw_rect.y + draw_rect.h); fflush(stdout);
			fixed_scale(image, scaled, pan_rect.x * scale_inv, (pan_rect.y + draw_rect.y) * scale_inv, &draw_rect, scale, scale_inv);
		}

		if SDL_MUSTLOCK(image) SDL_UnlockSurface(image);
		if SDL_MUSTLOCK(scaled) SDL_UnlockSurface(scaled);
	}

	drawn_rect = pan_rect;

	// swap width and height for 90 and 270
	if (rotation == 1 || rotation == 3) {
		rotated_rect.w = scaled_rect.h;
		rotated_rect.h = scaled_rect.w;
	} else {
		rotated_rect.w = scaled_rect.w;
		rotated_rect.h = scaled_rect.h;
	}

	// upper left corner of scaled image
	rotated_rect.x = (rotation == 1 || rotation == 2) ? scaled_rot[0]->w - rotated_rect.w : 0;
	rotated_rect.y = (rotation == 3 || rotation == 2) ? scaled_rot[0]->h - rotated_rect.h : 0;

	// center the scaled image when smaller than SCREEN_WIDTHxSCREEN_HEIGHT
	surface_rect.x = (scaled_rot[0]->w - rotated_rect.w) / 2;
	surface_rect.y = (scaled_rot[0]->h - rotated_rect.h) / 2;

	rotate();

	SDL_FillRect(surface, NULL, 0);
	SDL_BlitSurface(scaled_rot[0], &rotated_rect, surface, &surface_rect);
}

// kate: tab-width 4;
